from phylogenie.generators.dataset import (
    DatasetGenerator,
    DatasetGeneratorConfig,
    DatasetGeneratorRegistry,
)
from phylogenie.generators.msa import (
    MSAGenerator,
    MSAGeneratorConfig,
    MSAGeneratorRegistry,
)
from phylogenie.generators.tree import (
    TreeGenerator,
    TreeGeneratorConfig,
    TreeGeneratorRegistry,
)

__all__ = [
    "DatasetGenerator",
    "DatasetGeneratorConfig",
    "DatasetGeneratorRegistry",
    "MSAGenerator",
    "MSAGeneratorConfig",
    "MSAGeneratorRegistry",
    "TreeGenerator",
    "TreeGeneratorConfig",
    "TreeGeneratorRegistry",
]



from typing import Any

from numpy.random import Generator

import phylogenie.generators.configs as cfg
import phylogenie.generators.factories as f
import phylogenie.treesimulator.parameterizations.open_population as pmt
from phylogenie.generators.tree import TreeGeneratorRegistry
from phylogenie.plugins.native.base import PhylogenieTreeGenerator
from phylogenie.treesimulator import Model


@TreeGeneratorRegistry.register("phylogenie.canonical")
class CanonicalGenerator(PhylogenieTreeGenerator):
    states: list[str]
    init_state: str
    sampling_rates: cfg.SkylineVector = 0
    remove_after_sampling: bool = False
    birth_rates: cfg.SkylineVector = 0
    death_rates: cfg.SkylineVector = 0
    migration_rates: cfg.SkylineMatrix = None
    birth_rates_among_states: cfg.SkylineMatrix = None

    def _get_model(self, context: dict[str, Any], rng: Generator) -> Model:
        return pmt.get_canonical_model(
            init_state=self.init_state.format(**context),
            states=self.states,
            sampling_rates=f.skyline_vector(self.sampling_rates, context),
            remove_after_sampling=self.remove_after_sampling,
            birth_rates=f.skyline_vector(self.birth_rates, context),
            death_rates=f.skyline_vector(self.death_rates, context),
            migration_rates=f.skyline_matrix(self.migration_rates, context),
            birth_rates_among_states=f.skyline_matrix(
                self.birth_rates_among_states, context
            ),
        )


@TreeGeneratorRegistry.register("phylogenie.FBD")
class FBDGenerator(PhylogenieTreeGenerator):
    states: list[str]
    init_state: str
    sampling_proportions: cfg.SkylineVector = 0
    diversification: cfg.SkylineVector = 0
    turnover: cfg.SkylineVector = 0
    migration_rates: cfg.SkylineMatrix = None
    diversification_between_states: cfg.SkylineMatrix = None

    def _get_model(self, context: dict[str, Any], rng: Generator) -> Model:
        return pmt.get_FBD_model(
            init_state=self.init_state.format(**context),
            states=self.states,
            diversification=f.skyline_vector(self.diversification, context),
            turnover=f.skyline_vector(self.turnover, context),
            sampling_proportions=f.skyline_vector(self.sampling_proportions, context),
            migration_rates=f.skyline_matrix(self.migration_rates, context),
            diversification_between_states=f.skyline_matrix(
                self.diversification_between_states, context
            ),
        )


@TreeGeneratorRegistry.register("phylogenie.epidemiological")
class EpidemiologicalGenerator(PhylogenieTreeGenerator):
    states: list[str]
    init_state: str
    sampling_proportions: cfg.SkylineVector
    reproduction_numbers: cfg.SkylineVector = 0
    become_uninfectious_rates: cfg.SkylineVector = 0
    migration_rates: cfg.SkylineMatrix = None
    reproduction_numbers_among_states: cfg.SkylineMatrix = None

    def _get_model(self, context: dict[str, Any], rng: Generator) -> Model:
        return pmt.get_epidemiological_model(
            init_state=self.init_state.format(**context),
            states=self.states,
            reproduction_numbers=f.skyline_vector(self.reproduction_numbers, context),
            become_uninfectious_rates=f.skyline_vector(
                self.become_uninfectious_rates, context
            ),
            sampling_proportions=f.skyline_vector(self.sampling_proportions, context),
            migration_rates=f.skyline_matrix(self.migration_rates, context),
            reproduction_numbers_among_states=f.skyline_matrix(
                self.reproduction_numbers_among_states, context
            ),
        )


@TreeGeneratorRegistry.register("phylogenie.BD")
class BDGenerator(PhylogenieTreeGenerator):
    reproduction_number: cfg.SkylineParameter
    infectious_period: cfg.SkylineParameter
    sampling_proportion: cfg.SkylineParameter

    def _get_model(self, context: dict[str, Any], rng: Generator) -> Model:
        return pmt.get_BD_model(
            reproduction_number=f.skyline_parameter(self.reproduction_number, context),
            infectious_period=f.skyline_parameter(self.infectious_period, context),
            sampling_proportion=f.skyline_parameter(self.sampling_proportion, context),
        )


@TreeGeneratorRegistry.register("phylogenie.BDEI")
class BDEIGenerator(PhylogenieTreeGenerator):
    init_state: str = pmt.EXPOSED_STATE
    reproduction_number: cfg.SkylineParameter
    infectious_period: cfg.SkylineParameter
    incubation_period: cfg.SkylineParameter
    sampling_proportion: cfg.SkylineParameter

    def _get_model(self, context: dict[str, Any], rng: Generator) -> Model:
        init_state = self.init_state.format(**context)
        if init_state not in [pmt.EXPOSED_STATE, pmt.INFECTIOUS_STATE]:
            raise ValueError(
                f"Invalid init_state '{init_state}' for BDEI model. "
                f"It must be either '{pmt.EXPOSED_STATE}' or '{pmt.INFECTIOUS_STATE}'."
            )
        return pmt.get_BDEI_model(
            init_state=init_state,
            reproduction_number=f.skyline_parameter(self.reproduction_number, context),
            infectious_period=f.skyline_parameter(self.infectious_period, context),
            incubation_period=f.skyline_parameter(self.incubation_period, context),
            sampling_proportion=f.skyline_parameter(self.sampling_proportion, context),
        )


@TreeGeneratorRegistry.register("phylogenie.BDSS")
class BDSSGenerator(PhylogenieTreeGenerator):
    init_state: str = pmt.INFECTIOUS_STATE
    reproduction_number: cfg.SkylineParameter
    infectious_period: cfg.SkylineParameter
    superspreading_ratio: cfg.SkylineParameter
    superspreaders_proportion: cfg.SkylineParameter
    sampling_proportion: cfg.SkylineParameter

    def _get_model(self, context: dict[str, Any], rng: Generator) -> Model:
        init_state = self.init_state.format(**context)
        if init_state not in [pmt.INFECTIOUS_STATE, pmt.SUPERSPREADER_STATE]:
            raise ValueError(
                f"Invalid init_state '{init_state}' for BDSS model. "
                f"It must be either '{pmt.INFECTIOUS_STATE}' or '{pmt.SUPERSPREADER_STATE}'."
            )
        return pmt.get_BDSS_model(
            init_state=init_state,
            reproduction_number=f.skyline_parameter(self.reproduction_number, context),
            infectious_period=f.skyline_parameter(self.infectious_period, context),
            superspreading_ratio=f.skyline_parameter(
                self.superspreading_ratio, context
            ),
            superspreaders_proportion=f.skyline_parameter(
                self.superspreaders_proportion, context
            ),
            sampling_proportion=f.skyline_parameter(self.sampling_proportion, context),
        )



from phylogenie.skyline import SkylineParameterLike, skyline_parameter
from phylogenie.treesimulator.core import Model
from phylogenie.treesimulator.parameterizations.core import (
    Death,
    Sampling,
    StochasticEvent,
)

INFECTIOUS_STATE = "I"
SUSCEPTIBLE_STATE = "S"

SUSCEPTIBLES = "susceptibles"


class Transmission:
    def reactant_combinations(self, model: Model) -> int:
        return model[SUSCEPTIBLES] * model.count_active_nodes(INFECTIOUS_STATE)

    def apply(self, model: Model):
        model[SUSCEPTIBLES] -= 1
        parent_node = model.draw_active_node(INFECTIOUS_STATE)
        model.birth_from(parent_node, INFECTIOUS_STATE)


def get_SIR_model(
    transmission_rate: SkylineParameterLike,
    recovery_rate: SkylineParameterLike,
    sampling_rate: SkylineParameterLike,
    susceptibles: int,
):
    model = Model(
        init_state=INFECTIOUS_STATE, init_metadata={SUSCEPTIBLES: susceptibles}
    )
    model.add_event(
        StochasticEvent(rate=skyline_parameter(transmission_rate), fn=Transmission())
    )
    model.add_event(
        StochasticEvent(
            rate=skyline_parameter(recovery_rate), fn=Death(state=INFECTIOUS_STATE)
        )
    )
    model.add_event(
        StochasticEvent(
            rate=skyline_parameter(sampling_rate),
            fn=Sampling(state=INFECTIOUS_STATE, removal=True),
        )
    )
    return model



from phylogenie.treesimulator.parameterizations.core import (
    Death,
    Migration,
    Sampling,
    SingleReactantEventFunction,
    StochasticEvent,
    StochasticEventFunction,
    TimedEvent,
    TimedEventFunction,
)

__all__ = [
    "Death",
    "Migration",
    "Sampling",
    "SingleReactantEventFunction",
    "StochasticEvent",
    "StochasticEventFunction",
    "TimedEvent",
    "TimedEventFunction",
]



from phylogenie.draw import (
    draw_colored_dated_tree_categorical,
    draw_colored_dated_tree_continuous,
    draw_colored_tree_categorical,
    draw_colored_tree_continuous,
    draw_dated_tree,
    draw_tree,
)
from phylogenie.io import dump_fasta, dump_newick, load_fasta, load_newick, load_nexus
from phylogenie.main import run
from phylogenie.msa import MSA, Sequence
from phylogenie.skyline import (
    SkylineMatrix,
    SkylineMatrixCoercible,
    SkylineParameter,
    SkylineParameterLike,
    SkylineVector,
    SkylineVectorCoercible,
    SkylineVectorLike,
    skyline_matrix,
    skyline_parameter,
    skyline_vector,
)
from phylogenie.tree_node import TreeNode

__all__ = [
    "draw_colored_dated_tree_categorical",
    "draw_colored_dated_tree_continuous",
    "draw_colored_tree_categorical",
    "draw_colored_tree_continuous",
    "draw_dated_tree",
    "draw_tree",
    "dump_fasta",
    "dump_newick",
    "load_fasta",
    "load_newick",
    "load_nexus",
    "run",
    "MSA",
    "Sequence",
    "SkylineMatrix",
    "SkylineMatrixCoercible",
    "SkylineParameter",
    "SkylineParameterLike",
    "SkylineVector",
    "SkylineVectorCoercible",
    "SkylineVectorLike",
    "skyline_matrix",
    "skyline_parameter",
    "skyline_vector",
    "TreeNode",
]



import os
from abc import ABC, abstractmethod
from typing import Annotated, Any

import phylogenie.generators.configs as cfg
from phylogenie.generators.dataset import DatasetGenerator, DatasetGeneratorRegistry
from phylogenie.utils import Registry


class TreeGenerator(ABC, cfg.StrictBaseModel):
    @abstractmethod
    def generate(
        self, filename: str, context: dict[str, Any], seed: int | None = None
    ) -> dict[str, Any]: ...


TreeGeneratorRegistry: Registry[TreeGenerator] = Registry(TreeGenerator)
TreeGeneratorConfig = Annotated[TreeGenerator, TreeGeneratorRegistry.Validator]


@DatasetGeneratorRegistry.register("tree")
class TreeDatasetGenerator(DatasetGenerator):
    tree_generator: TreeGeneratorConfig

    def _generate_from_context(
        self, data_dir: str, file_id: str, context: dict[str, Any], seed: int
    ) -> dict[str, Any]:
        return self.tree_generator.generate(
            os.path.join(data_dir, file_id), context, seed
        )



from abc import ABC, abstractmethod
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Protocol

from phylogenie.skyline import SkylineParameter
from phylogenie.tree_node import TreeNode
from phylogenie.treesimulator.core import STATE, Event, Model


def _get_next_greater(times: Iterable[float], current_time: float) -> float | None:
    return next((v for v in times if v > current_time), None)


class StochasticEventFunction(Protocol):
    def reactant_combinations(self, model: Model) -> int: ...
    def apply(self, model: Model): ...


class TimedEventFunction(Protocol):
    def max_firings(self, model: Model) -> int: ...
    def apply_firings(self, model: Model, firings: int): ...


@dataclass
class StochasticEvent(Event):
    rate: SkylineParameter
    fn: StochasticEventFunction

    def apply(self, model: Model):
        if model.current_time not in self.rate.change_times:
            self.fn.apply(model)

    def get_next_firing_time(self, model: Model) -> float | None:
        next_change_time = _get_next_greater(self.rate.change_times, model.current_time)
        rate = self.rate.get_value_at_time(model.current_time)
        propensity = rate * self.fn.reactant_combinations(model)
        if not propensity:
            return next_change_time
        next_firing_time = model.current_time + model.rng.expovariate(propensity)
        return (
            next_firing_time
            if next_change_time is None
            else min(next_firing_time, next_change_time)
        )


@dataclass
class TimedEvent(Event):
    times: Iterable[float]
    firings: float | int
    fn: TimedEventFunction

    def apply(self, model: Model):
        max_firings = self.fn.max_firings(model)
        firings = (
            int(self.firings * max_firings)
            if isinstance(self.firings, float)
            else min(self.firings, max_firings)
        )
        self.fn.apply_firings(model, firings)

    def get_next_firing_time(self, model: Model) -> float | None:
        return _get_next_greater(self.times, model.current_time)


@dataclass(kw_only=True)
class SingleReactantEventFunction(ABC):
    state: str | None = None

    def draw_one(self, model: Model) -> TreeNode:
        return model.draw_active_node(self.state)

    def reactant_combinations(self, model: Model) -> int:
        return model.count_active_nodes(self.state)

    @abstractmethod
    def apply_to_node(self, model: Model, node: TreeNode): ...

    def max_firings(self, model: Model) -> int:
        return model.count_active_nodes(self.state)

    def apply_firings(self, model: Model, firings: int):
        active_nodes = model.get_active_nodes(self.state)
        for node in model.rng.sample(active_nodes, firings):
            self.apply_to_node(model, node)

    def apply(self, model: Model):
        self.apply_to_node(model, self.draw_one(model))


class Death(SingleReactantEventFunction):
    def apply_to_node(self, model: Model, node: TreeNode):
        return model.remove(node)


@dataclass(kw_only=True)
class Migration(SingleReactantEventFunction):
    target_state: str

    def apply_to_node(self, model: Model, node: TreeNode):
        model.migrate(node, self.target_state)


@dataclass(kw_only=True)
class Sampling(SingleReactantEventFunction):
    removal: bool

    def apply_to_node(self, model: Model, node: TreeNode):
        if self.removal:
            model.sample(node)
        else:
            _, sample_node = model.birth_from(node, node[STATE])
            model.sample(sample_node)



import subprocess
from typing import Any

import phylogenie.generators.factories as f
from phylogenie.generators.msa import MSAGenerator, MSAGeneratorRegistry


@MSAGeneratorRegistry.register("alisim")
class AliSimGenerator(MSAGenerator):
    iqtree_path: str = "iqtree2"
    args: dict[str, Any]

    def generate(
        self,
        filename: str,
        input_tree_file: str,
        context: dict[str, Any],
        seed: int | None = None,
    ) -> None:
        command = [
            self.iqtree_path,
            "--alisim",
            filename,
            "--tree",
            input_tree_file,
            "--seed",
            str(seed),
        ]

        for key, value in self.args.items():
            command.extend([key, f.string(value, context)])

        command.extend(["-af", "fasta"])
        subprocess.run(command, check=True, stdout=subprocess.DEVNULL)
        subprocess.run(["rm", f"{input_tree_file}.log"], check=True)



import re
from collections import defaultdict
from random import Random
from typing import Any, Protocol

from phylogenie.tree_node import TreeNode
from phylogenie.utils import MetadataMixin, OrderedSet

STATE = "state"


class Event(Protocol):
    def get_next_firing_time(self, model: "Model") -> float | None: ...
    def apply(self, model: "Model") -> None: ...


class Model(MetadataMixin):
    def __init__(self, init_state: str, init_metadata: dict[str, Any] | None = None):
        super().__init__()
        self.rng = Random()
        self._init_state = init_state
        self._init_metadata = init_metadata
        self._events: list[Event] = []

    @staticmethod
    def _get_node_name(node_id: int, state: str) -> str:
        return f"{node_id}|{state}"

    def reset(self):
        self.clear()
        if self._init_metadata is not None:
            self._metadata = self._init_metadata.copy()
        self._current_time = 0.0
        self._next_node_id = 0
        self._active_nodes: dict[str, OrderedSet[TreeNode]] = defaultdict(OrderedSet)
        self._sampled_nodes: set[str] = set()
        self._node_times: dict[TreeNode, float] = {}
        self._tree = self._get_new_node(self._init_state)
        self._run_events = self._events.copy()

    @property
    def current_time(self) -> float:
        return self._current_time

    def _get_new_node(self, state: str) -> TreeNode:
        self._next_node_id += 1
        node = TreeNode(self._get_node_name(self._next_node_id, state))
        node[STATE] = state
        self._active_nodes[state].add(node)
        return node

    def _fix(self, node: TreeNode):
        if node.branch_length is not None:
            raise RuntimeError(f"Node {node} has already been fixed")
        parent_time = 0.0 if node.parent is None else self._node_times[node.parent]
        node.branch_length = self.current_time - parent_time
        self._node_times[node] = self.current_time
        self._active_nodes[node[STATE]].remove(node)

    def _stem(self, node: TreeNode, stem_state: str) -> TreeNode:
        self._fix(node)
        stem_node = self._get_new_node(stem_state)
        node.add_child(stem_node)
        return stem_node

    def remove(self, node: TreeNode):
        self._fix(node)

    def migrate(self, node: TreeNode, new_state: str) -> TreeNode:
        return self._stem(node, new_state)

    def birth_from(self, parent: TreeNode, new_state: str) -> tuple[TreeNode, TreeNode]:
        new_node = self._get_new_node(new_state)
        parent.add_child(new_node)
        stem_node = self._stem(parent, parent[STATE])
        return stem_node, new_node

    def sample(self, node: TreeNode):
        self._sampled_nodes.add(node.name)
        self._fix(node)

    def get_sampled_tree(self) -> TreeNode | None:
        tree = self._tree.copy()
        for node in list(tree.iter_postorder()):
            if node.name not in self._sampled_nodes and not node.children:
                if node.parent is None:
                    return None
                else:
                    node.parent.remove_child(node)
            elif len(node.children) == 1:
                (child,) = node.children
                child.update_parent(node.parent)
                child.branch_length += node.branch_length  # pyright: ignore
                if node.parent is None:
                    return child
                else:
                    node.parent.remove_child(node)
        return tree

    @property
    def n_sampled(self) -> int:
        return len(self._sampled_nodes)

    def get_active_nodes(self, state: str | None = None) -> tuple[TreeNode, ...]:
        return tuple(
            node
            for s, nodes in self._active_nodes.items()
            if state is None or re.fullmatch(state, s) is not None
            for node in nodes
        )

    def count_active_nodes(self, state: str | None = None) -> int:
        return len(self.get_active_nodes(state))

    def draw_active_node(self, state: str | None = None) -> TreeNode:
        return self.rng.choice(self.get_active_nodes(state))

    def step(self, max_time: float | None = None) -> bool:
        next_event_times = [e.get_next_firing_time(self) for e in self._run_events]
        next_step_times = [t for t in [*next_event_times, max_time] if t is not None]
        if not next_step_times:
            return False

        self._current_time = min(next_step_times)
        for event, time in zip(self._run_events, next_event_times):
            if time == self._current_time:
                event.apply(self)

        return self._current_time != max_time

    def add_event(self, event: Event):
        self._events.append(event)

    def add_run_event(self, event: Event):
        self._run_events.append(event)



import math
from collections import deque
from collections.abc import Iterator, Mapping
from typing import Any

from phylogenie.utils import MetadataMixin


class TreeNode(MetadataMixin):
    """
    Phylogenetic tree node with parent-child relationships and metadata.

    This class represents both a node in a phylogenetic tree and the subtree
    rooted at that node. It supports parent-child relationships, tree traversal,
    various tree properties and metrics, and metadata annotations.
    """

    def __init__(self, name: str = "", branch_length: float | None = None):
        """Initialize a tree node with optional name and branch length."""
        super().__init__()
        self.name = name
        self.branch_length = branch_length
        self._parent: TreeNode | None = None
        self._children: list[TreeNode] = []

    # ----------------
    # Basic properties
    # ----------------
    # Properties related to parent-child relationships.

    @property
    def children(self) -> tuple["TreeNode", ...]:
        """Return the children of this node as a tuple."""
        return tuple(self._children)

    @property
    def parent(self) -> "TreeNode | None":
        """Return the parent of this node, if any."""
        return self._parent

    @property
    def root(self) -> "TreeNode":
        """Return the root node of the tree containing this node."""
        node = self
        while node.parent is not None:
            node = node.parent
        return node

    def add_child(self, child: "TreeNode") -> "TreeNode":
        """
        Attach a child node to this node.

        This updates both the child's parent pointer and this node's internal
        children list. A child may only have one parent.
        """
        if child.parent is not None:
            raise ValueError(f"Node {child.name} already has a parent.")
        child._parent = self
        self._children.append(child)
        return self

    def remove_child(self, child: "TreeNode"):
        """
        Detach a child node from this node.

        The child is removed from the children list and its parent pointer is
        cleared. The child must already be attached to this node.
        """
        if child not in self._children:
            raise ValueError(f"Node {child.name} is not a child of node {self.name}.")
        self._children.remove(child)
        child._parent = None

    def update_parent(self, parent: "TreeNode | None"):
        """
        Set or clear this node's parent, updating both sides of the link.

        If the node currently has a parent, it is removed from that parent's
        children list before the new relationship is created.
        """
        if self.parent is not None:
            self.parent.remove_child(self)
        self._parent = parent
        if parent is not None:
            parent._children.append(self)

    def is_leaf(self) -> bool:
        """Determine whether this node is a leaf (has no children)."""
        return not self.children

    def is_internal(self) -> bool:
        """Determine whether this node is internal (has one or more children)."""
        return not self.is_leaf()

    def get_leaves(self) -> tuple["TreeNode", ...]:
        """Return all leaf nodes in the subtree rooted at this node."""
        return tuple(node for node in self if node.is_leaf())

    def get_internal_nodes(self) -> tuple["TreeNode", ...]:
        """Return all internal (non-leaf) nodes in the subtree rooted at this node."""
        return tuple(node for node in self if node.is_internal())

    def is_binary(self) -> bool:
        """Determine whether the subtree rooted at this node is binary (every node has 0 or 2 children)."""
        return all(len(node.children) == 2 for node in self.get_internal_nodes())

    def branch_length_or_raise(self) -> float:
        """
        Return the branch length, raising if it is missing.

        The root node returns 0.0 when its branch length is None to provide a
        convenient origin length for distance calculations.
        """
        if self.parent is None:
            return 0 if self.branch_length is None else self.branch_length
        if self.branch_length is None:
            raise ValueError(f"Branch length of node {self.name} is not set.")
        return self.branch_length

    # --------------
    # Tree traversal
    # --------------
    # Methods for traversing the tree in various orders.

    def iter_ancestors(self, stop: "TreeNode | None" = None) -> Iterator["TreeNode"]:
        """Iterate over ancestors from the parent up to (but excluding) a stop node."""
        node = self
        while True:
            if node.parent is None:
                if stop is None:
                    return
                raise ValueError("Reached root without encountering stop node.")
            node = node.parent
            if node == stop:
                return
            yield node

    def iter_upward(self, stop: "TreeNode | None" = None) -> Iterator["TreeNode"]:
        """Iterate from this node upward through its ancestors, including self and up to (but excluding) a stop node."""
        if self == stop:
            return
        yield self
        yield from self.iter_ancestors(stop=stop)

    def iter_descendants(self) -> Iterator["TreeNode"]:
        """Iterate over all descendants of this node (excluding self)."""
        for child in self.children:
            yield child
            yield from child.iter_descendants()

    def iter_preorder(self) -> Iterator["TreeNode"]:
        """Iterate over nodes in preorder (self before descendants)."""
        yield self
        yield from self.iter_descendants()

    def iter_inorder(self) -> Iterator["TreeNode"]:
        """
        Iterate over nodes in inorder for binary trees.

        For non-binary trees, a ValueError is raised because inorder is not
        well-defined.
        """
        if self.is_leaf():
            yield self
            return

        if len(self.children) != 2:
            raise ValueError("Inorder traversal is only defined for binary trees.")

        left, right = self.children
        yield from left.iter_inorder()
        yield self
        yield from right.iter_inorder()

    def iter_postorder(self) -> Iterator["TreeNode"]:
        """Iterate over nodes in postorder (descendants before self)."""
        for child in self.children:
            yield from child.iter_postorder()
        yield self

    def iter_breadth_first(self) -> Iterator["TreeNode"]:
        """Iterate over nodes in breadth-first order."""
        queue: deque["TreeNode"] = deque([self])
        while queue:
            node = queue.popleft()
            yield node
            queue.extend(node.children)

    # --------------
    # Paths and hops
    # --------------

    def get_mrca(self, other: "TreeNode") -> "TreeNode":
        """Find the most recent common ancestor (MRCA) of this node and another node."""
        self_ancestors = set(self.iter_upward())
        for other_ancestor in other.iter_upward():
            if other_ancestor in self_ancestors:
                return other_ancestor
        raise ValueError(
            f"No common ancestor found between node {self.name} and node {other.name}."
        )

    def get_path(self, other: "TreeNode") -> list["TreeNode"]:
        """
        Get the node-by-node path between this node and another node.

        The returned list starts at this node, travels up to the MRCA, and then
        down to the target node.
        """
        mrca = self.get_mrca(other)
        return [
            *self.iter_upward(stop=mrca.parent),
            *reversed(list(other.iter_upward(stop=mrca))),
        ]

    def count_hops(self, other: "TreeNode") -> int:
        """Count the number of edges between this node and another node."""
        return len(self.get_path(other)) - 1

    def get_distance(self, other: "TreeNode") -> float:
        """Compute the branch-length distance between this node and another node."""
        mrca = self.get_mrca(other)
        path = self.get_path(other)
        path.remove(mrca)
        return sum(node.branch_length_or_raise() for node in path)

    # ---------------
    # Tree properties
    # ---------------

    @property
    def leaf_counts(self) -> dict["TreeNode", int]:
        """Compute the number of leaves under each node for the subtree rooted at this node."""
        n_leaves: dict[TreeNode, int] = {}
        for node in self.iter_postorder():
            n_leaves[node] = sum(n_leaves[child] for child in node.children) or 1
        return n_leaves

    @property
    def n_leaves(self) -> int:
        """Return the number of leaf nodes for the subtree rooted at this node."""
        return len(self.get_leaves())

    @property
    def height_levels(self) -> dict["TreeNode", int]:
        """
        Compute the height level of all nodes for the subtree rooted at this node.

        The height level is defined as the number of edges from a node to its farthest leaf.
        It is 0 for leaf nodes, and increases by 1 for each level up the tree.
        """
        height_levels: dict[TreeNode, int] = {}
        for node in self.iter_postorder():
            height_levels[node] = (
                0
                if node.is_leaf()
                else max(1 + height_levels[child] for child in node.children)
            )
        return height_levels

    @property
    def height_level(self) -> int:
        """
        Return the height level of the node.

        The height level is defined as the number of edges from a node to its farthest leaf.
        It is 0 for leaf nodes, and increases by 1 for each level up the tree.
        """
        return self.height_levels[self]

    @property
    def heights(self) -> dict["TreeNode", float]:
        """
        Compute the height of all nodes for the subtree rooted at this node.

        The height is defined as the distance in branch-length units from a node to its farthest leaf.
        It is 0 for leaf nodes, and increases by the branch lengths up the tree.
        """
        heights: dict["TreeNode", float] = {}
        for node in self.iter_postorder():
            heights[node] = (
                0
                if node.is_leaf()
                else max(
                    child.branch_length_or_raise() + heights[child]
                    for child in node.children
                )
            )
        return heights

    @property
    def height(self) -> float:
        """
        Return the height of the node.

        The height is defined as the distance in branch-length units from a node to its farthest leaf.
        It is 0 for leaf nodes, and increases by the branch lengths up the tree.
        """
        return self.heights[self]

    @property
    def depth_levels(self) -> dict["TreeNode", int]:
        """
        Compute the depth level of all nodes for the subtree rooted at this node.

        The depth level is defined as the number of edges from a node to the current node.
        It is 0 for the current node, and increases by 1 for each level down the tree.
        """
        depth_levels: dict[TreeNode, int] = {self: 0}
        for node in self.iter_descendants():
            depth_levels[node] = depth_levels[node.parent] + 1  # pyright: ignore
        return depth_levels

    @property
    def depth_level(self) -> int:
        """
        Return the depth level of the node.

        The depth level is defined as the number of edges from a node to the root.
        It is 0 for the root, and increases by 1 for each level down the tree.
        """
        return self.count_hops(self.root)

    @property
    def depths(self) -> dict["TreeNode", float]:
        """
        Compute the depth of all nodes for the subtree rooted at this node.

        The depth is defined as the distance in branch-length units from a node to the current node.
        It is 0 for the current node, and increases by the branch lengths down the tree.
        """
        depths: dict[TreeNode, float] = {self: 0.0}
        for node in self.iter_descendants():
            parent_depth = depths[node.parent]  # pyright: ignore
            depths[node] = node.branch_length_or_raise() + parent_depth
        return depths

    @property
    def depth(self) -> float:
        """
        Compute the depth of this node.

        The depth is defined as the distance in branch-length units from a node to the root.
        It is 0 for the root, and increases by the branch lengths down the tree.
        """
        return self.get_distance(self.root)

    @property
    def times(self) -> dict["TreeNode", float]:
        """
        Compute the time of all nodes for the subtree rooted at this node.

        Times are measured forwards from the current node.
        The time of the current node is defined as its branch length,
        and increases by the branch lengths down the tree.
        """
        origin = self.branch_length_or_raise()
        return {node: origin + depth for node, depth in self.depths.items()}

    @property
    def time(self) -> float:
        """
        Compute the time of this node.

        Times are measured forwards from the root of the tree.
        The time of the root is defined as its branch length,
        and increases by the branch lengths down the tree.

        Returns
        --------
        float
            The time of this node.
        """
        root = self.root
        origin = root.branch_length_or_raise()
        return self.get_distance(root) + origin

    @property
    def ages(self) -> dict["TreeNode", float]:
        """
        Compute the age of all nodes for the subtree rooted at this node.

        Ages are measured backwards in time from the most recent leaf of the subtree in branch-length units.
        The age of the most recent leaf is defined as 0, and increases by the branch lengths up the tree.
        """
        ages: dict[TreeNode, float] = {self: self.height}
        for node in self.iter_descendants():
            ages[node] = ages[node.parent] - node.branch_length_or_raise()  # pyright: ignore
        return ages

    @property
    def age(self) -> float:
        """
        Return the age of this node.

        Ages are measured backwards in time from the most recent leaf of the tree in branch-length units.
        The age of the most recent leaf is defined as 0, and increases by the branch lengths up the tree.
        """
        return self.root.height - self.get_distance(self.root)

    @property
    def origin(self) -> float:
        """
        Return the origin time of the subtree rooted at this node.

        The origin time is defined as the age of the current node plus its branch length.
        """
        return self.age + self.branch_length_or_raise()

    def compute_sackin_index(self, normalize: bool = False) -> float:
        """
        Compute the Sackin index for the subtree rooted at this node.

        Parameters
        -----------
        normalize : bool, optional
            If the normalize flag is set to True, the Sackin index is normalized to the range [0, 1] for binary trees,
            where 0 corresponds to a perfectly balanced tree and 1 corresponds to a completely unbalanced (pectinate) tree.
            For non-binary trees, normalization is not defined and a ValueError is raised. Default is False.
        """
        sackin_index = sum(
            dl for node, dl in self.depth_levels.items() if node.is_leaf()
        )
        if normalize:
            if not self.is_binary():
                raise ValueError(
                    "Normalized Sackin index is only defined for binary trees."
                )
            n = self.n_leaves
            h = math.floor(math.log2(n))
            min_sackin_index = n * (h + 2) - 2 ** (h + 1)
            max_sackin_index = n * (n - 1) / 2
            return (sackin_index - min_sackin_index) / (
                max_sackin_index - min_sackin_index
            )
        return sackin_index

    # -------------
    # Miscellaneous
    # -------------
    # Miscellaneous methods like copying and representation.

    def copy(self) -> "TreeNode":
        """
        Deep-copy this node and its descendants.

        The returned node is a new tree with duplicated metadata and structure.
        """
        new_node = TreeNode(self.name, self.branch_length)
        new_node.update(self.metadata)
        for child in self.children:
            new_node.add_child(child.copy())
        return new_node

    def ladderize(self, key: Mapping["TreeNode", Any] | None = None):
        """
        Sort children of all nodes to produce a ladderized tree.

        Parameters
        -----------
        key : Mapping["TreeNode", Any] | None, optional
            Mapping used to sort children. Defaults to the number of leaf descendants of each node.
        """
        if key is None:
            key = self.leaf_counts

        for node in self:
            node._children.sort(key=lambda child: key[child])

    def get_descendant(self, name: str) -> "TreeNode":
        """Find the first descendant node with the given name."""
        for node in self:
            if node.name == name:
                return node
        raise ValueError(f"Node {name} not found.")

    # -------------
    # Dunder methods
    # -------------

    def __repr__(self) -> str:
        """Return a string representation of the node."""
        return f"TreeNode(name='{self.name}', branch_length={self.branch_length}, metadata={self.metadata})"

    def __iter__(self) -> Iterator["TreeNode"]:
        """Iterate over all nodes in the subtree rooted at this node in preorder."""
        return self.iter_preorder()

    def __len__(self) -> int:
        """Return the number of nodes in the subtree rooted at this node."""
        return sum(1 for _ in self)



import os
from abc import ABC, abstractmethod
from typing import Annotated, Any

import phylogenie.generators.configs as cfg
from phylogenie.generators.dataset import DatasetGenerator, DatasetGeneratorRegistry
from phylogenie.generators.tree import TreeGeneratorConfig
from phylogenie.io import dump_newick, load_newick
from phylogenie.tree_node import TreeNode
from phylogenie.utils import Registry

MSAS_DIRNAME = "MSAs"
TREES_DIRNAME = "trees"


class MSAGenerator(ABC, cfg.StrictBaseModel):
    @abstractmethod
    def generate(
        self,
        filename: str,
        input_tree_file: str,
        context: dict[str, Any],
        seed: int | None = None,
    ) -> None: ...


MSAGeneratorRegistry: Registry["MSAGenerator"] = Registry(MSAGenerator)
MSAGeneratorConfig = Annotated[MSAGenerator, MSAGeneratorRegistry.Validator]


@DatasetGeneratorRegistry.register("msa")
class MSADatasetGenerator(DatasetGenerator):
    tree_generator: TreeGeneratorConfig
    msa_generator: MSAGeneratorConfig
    keep_trees: bool = False

    def _generate_from_context(
        self, data_dir: str, file_id: str, context: dict[str, Any], seed: int
    ) -> dict[str, Any]:
        trees_dir = (
            os.path.join(data_dir, TREES_DIRNAME) if self.keep_trees else data_dir
        )
        msas_dir = os.path.join(data_dir, MSAS_DIRNAME) if self.keep_trees else data_dir

        os.makedirs(trees_dir, exist_ok=True)
        os.makedirs(msas_dir, exist_ok=True)

        tree_filename = os.path.join(trees_dir, file_id)
        msa_filename = os.path.join(msas_dir, file_id)

        metadata = self.tree_generator.generate(tree_filename, context, seed)

        tree_file = f"{tree_filename}.nwk"
        tree = load_newick(tree_file)
        assert isinstance(tree, TreeNode)

        times = tree.times
        for leaf in tree.get_leaves():
            leaf.name += f"|{times[leaf]}"
        dump_newick(tree, tree_file)

        self.msa_generator.generate(msa_filename, tree_file, context, seed)
        if not self.keep_trees:
            os.remove(tree_file)

        return metadata



"""
Core configuration models.

This module provides foundational configuration models used as building
blocks for generator configurations. Each configuration describes the
inputs required by a factory function to construct a specific component.

The models defined here standardize how scalar values, distributions,
and time-varying ("skyline") parameters are represented, enabling
flexible and composable configuration schemas across the codebase.
"""

from typing import Any

from pydantic import BaseModel, ConfigDict

import phylogenie.typings as pgt


class StrictBaseModel(BaseModel):
    model_config = ConfigDict(extra="forbid")


class Distribution(BaseModel):
    type: str
    model_config = ConfigDict(extra="allow")

    @property
    def args(self) -> dict[str, Any]:
        return self.model_extra  # pyright: ignore


Context = dict[str, str | Distribution]
Integer = str | int
Scalar = str | pgt.Scalar
ManyScalars = str | pgt.Many[Scalar]
OneOrManyScalars = Scalar | pgt.Many[Scalar]
OneOrMany2DScalars = Scalar | pgt.Many2D[Scalar]


class SkylineParameterModel(StrictBaseModel):
    value: ManyScalars
    change_times: ManyScalars


class SkylineVectorModel(StrictBaseModel):
    value: str | pgt.Many[OneOrManyScalars]
    change_times: ManyScalars


class SkylineMatrixModel(StrictBaseModel):
    value: str | pgt.Many[OneOrMany2DScalars]
    change_times: ManyScalars


SkylineParameter = Scalar | SkylineParameterModel
SkylineVector = str | pgt.Scalar | pgt.Many[SkylineParameter] | SkylineVectorModel
SkylineMatrix = str | pgt.Scalar | pgt.Many[SkylineVector] | SkylineMatrixModel | None



from abc import abstractmethod
from typing import Annotated, Any, Callable

from numpy.random import Generator, default_rng
from pydantic import Field

import phylogenie.generators.configs as cfg
import phylogenie.generators.factories as f
from phylogenie.generators.tree import TreeGenerator
from phylogenie.io import dump_newick
from phylogenie.plugins.native.timed_events import TimedEventConfig
from phylogenie.tree_node import TreeNode
from phylogenie.treesimulator import Model, simulate_tree


class PhylogenieTreeGenerator(TreeGenerator):
    n_leaves: cfg.Integer | None = None
    max_time: cfg.Scalar | None = None
    timeout: float | None = None
    acceptance_criterion: str | None = None
    tree_logs: Annotated[dict[str, str], Field(default_factory=dict)]
    model_logs: Annotated[dict[str, str], Field(default_factory=dict)]
    timed_events: Annotated[list[TimedEventConfig], Field(default_factory=list)]

    @abstractmethod
    def _get_model(self, context: dict[str, Any], rng: Generator) -> Model: ...

    def generate(
        self, filename: str, context: dict[str, Any], seed: int | None = None
    ) -> dict[str, Any]:
        rng = default_rng(seed)
        model = self._get_model(context, rng)
        model.rng.seed(seed)

        for event in self.timed_events:
            model.add_event(event.factory(context, rng))

        max_time = None if self.max_time is None else f.scalar(self.max_time, context)
        n_leaves = None if self.n_leaves is None else f.integer(self.n_leaves, context)

        acceptance_criterion: None | Callable[[TreeNode], bool] = (
            None
            if self.acceptance_criterion is None
            else lambda tree: f.eval_expression(
                self.acceptance_criterion,  # pyright: ignore
                context,
                extra_context={"tree": tree},
            )
        )

        def _tree_logs(tree: TreeNode) -> dict[str, Any]:
            return {
                key: f.eval_expression(expr, context, extra_context={"tree": tree})
                for key, expr in self.tree_logs.items()
            }

        def _model_logs(model: Model) -> dict[str, Any]:
            return {
                key: f.eval_expression(expr, context, extra_context={"model": model})
                for key, expr in self.model_logs.items()
            }

        tree, metadata = simulate_tree(
            model=model,
            n_leaves=n_leaves,
            max_time=max_time,
            timeout=self.timeout,
            acceptance_criterion=acceptance_criterion,
            tree_logs=_tree_logs,
            model_logs=_model_logs,
        )

        dump_newick(tree, f"{filename}.nwk")
        return metadata



from typing import Any

from numpy.random import Generator

import phylogenie.generators.configs as cfg
import phylogenie.generators.factories as f
import phylogenie.treesimulator.parameterizations.sir as pmt
from phylogenie.generators.tree import TreeGeneratorRegistry
from phylogenie.plugins.native.base import PhylogenieTreeGenerator
from phylogenie.treesimulator import Model


@TreeGeneratorRegistry.register("phylogenie.SIR")
class SIRGenerator(PhylogenieTreeGenerator):
    transmission_rate: cfg.SkylineParameter
    recovery_rate: cfg.SkylineParameter
    sampling_rate: cfg.SkylineParameter
    susceptibles: cfg.Integer

    def _get_model(self, context: dict[str, Any], rng: Generator) -> Model:
        return pmt.get_SIR_model(
            transmission_rate=f.skyline_parameter(self.transmission_rate, context),
            recovery_rate=f.skyline_parameter(self.recovery_rate, context),
            sampling_rate=f.skyline_parameter(self.sampling_rate, context),
            susceptibles=f.integer(self.susceptibles, context),
        )



import os
from argparse import ArgumentParser
from glob import glob
from importlib import import_module
from importlib.metadata import entry_points

from pydantic import TypeAdapter, ValidationError
from yaml import safe_load

from phylogenie.generators import DatasetGeneratorConfig
from phylogenie.generators.dataset import DatasetGenerator


def _format_validation_error(e: ValidationError) -> str:
    """Format a Pydantic ValidationError for console output."""
    formatted_errors = [
        f"- {'.'.join(str(loc) for loc in err['loc'])}: {err['msg']} ({err['type']})"
        for err in e.errors()
    ]
    return "\n".join(formatted_errors)


def _generate_from_config_file(config_file: str):
    """Load a configuration file, validate it, and run generation."""
    adapter: TypeAdapter[DatasetGenerator] = TypeAdapter(DatasetGeneratorConfig)
    try:
        with open(config_file, "r") as f:
            try:
                config = safe_load(f)
            except Exception as e:
                print(f"❌ Failed to parse {config_file}: {e}")
                exit(-1)

            try:
                generator = adapter.validate_python(config)
            except ValidationError as e:
                print("❌ Invalid configuration:")
                print(_format_validation_error(e))
                exit(-1)
    except Exception as e:
        print(f"❌ Failed to load {config_file}: {e}")
        exit(-1)
    generator.generate()


def run(config_path: str):
    """Run dataset generation from a config file or directory."""
    for ep in entry_points(group="phylogenie.plugins"):
        import_module(ep.value)

    if os.path.isdir(config_path):
        for config_file in glob(os.path.join(config_path, "**/*.yaml"), recursive=True):
            _generate_from_config_file(config_file)
    else:
        _generate_from_config_file(config_path)


def main():
    """Entry point for the CLI."""
    parser = ArgumentParser(
        description="Generate dataset(s) starting from provided config(s)."
    )
    parser.add_argument(
        "config_path",
        type=str,
        help="Path to a config file or a directory containing config files.",
    )
    args = parser.parse_args()

    run(args.config_path)



from phylogenie.plugins.native.base import PhylogenieTreeGenerator
from phylogenie.plugins.native.timed_events import (
    TimedEvent,
    TimedEventConfig,
    TimedEventRegistry,
)

__all__ = [
    "PhylogenieTreeGenerator",
    "TimedEvent",
    "TimedEventConfig",
    "TimedEventRegistry",
]



from abc import ABC, abstractmethod
from typing import Annotated, Any

from numpy.random import Generator
from pydantic import BeforeValidator

import phylogenie.generators.configs as cfg
import phylogenie.generators.factories as f
import phylogenie.treesimulator.parameterizations as pmt
from phylogenie.utils import Registry


class TimedEvent(ABC, cfg.StrictBaseModel):
    times: cfg.ManyScalars
    firings: cfg.Scalar

    @abstractmethod
    def fn_factory(
        self, context: dict[str, Any], rng: Generator
    ) -> pmt.TimedEventFunction: ...

    def factory(self, context: dict[str, Any], rng: Generator) -> pmt.TimedEvent:
        times = f.many_scalars(self.times, context)
        firings = f.scalar(self.firings, context)
        return pmt.TimedEvent(
            times=times, firings=firings, fn=self.fn_factory(context, rng)
        )


TimedEventRegistry = Registry(TimedEvent)


@TimedEventRegistry.register("sampling")
class TimedSamplingModel(TimedEvent):
    state: str | None = None
    removal: bool

    def fn_factory(self, context: dict[str, Any], rng: Generator):
        return pmt.Sampling(state=self.state, removal=self.removal)


@TimedEventRegistry.register("death")
class TimedDeathModel(TimedEvent):
    state: str | None = None

    def fn_factory(self, context: dict[str, Any], rng: Generator):
        return pmt.Death(state=self.state)


TimedEventConfig = Annotated[
    TimedEvent, BeforeValidator(lambda v: TimedEventRegistry.load(v))
]



from collections.abc import Callable, Iterator, Mapping
from types import MappingProxyType
from typing import Any, Generic, TypeVar

from pydantic import BaseModel, BeforeValidator

import phylogenie.typeguards as tg

T = TypeVar("T")
BaseModelT = TypeVar("BaseModelT", bound=BaseModel)


class OrderedSet(Generic[T]):
    def __init__(self):
        self._dict: dict[T, None] = {}

    def add(self, item: T):
        self._dict[item] = None

    def remove(self, item: T):
        del self._dict[item]

    def __contains__(self, item: T) -> bool:
        return item in self._dict

    def __iter__(self) -> Iterator[T]:
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)


class MetadataMixin:
    def __init__(self):
        self._metadata: dict[str, Any] = {}

    @property
    def metadata(self) -> Mapping[str, Any]:
        return MappingProxyType(self._metadata)

    def set(self, key: str, value: Any):
        self._metadata[key] = value

    def update(self, metadata: Mapping[str, Any]):
        self._metadata.update(metadata)

    def get(self, key: str, default: Any = None) -> Any:
        return self._metadata.get(key, default)

    def delete(self, key: str):
        del self._metadata[key]

    def clear(self):
        self._metadata.clear()

    def __getitem__(self, key: str) -> Any:
        return self._metadata[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self._metadata[key] = value


class Registry(Generic[BaseModelT]):
    def __init__(self, base_type: type[BaseModelT], discriminator: str = "spec"):
        self.name = base_type.__name__
        self.base_type = base_type
        self.discriminator = discriminator
        self._registry: dict[str, type[BaseModelT]] = {}

    def register(self, key: str) -> Callable[[type[BaseModelT]], type[BaseModelT]]:
        def decorator(cls: type[BaseModelT]) -> type[BaseModelT]:
            if not issubclass(cls, self.base_type):
                raise TypeError(
                    f"{cls.__name__} must be a subclass of {self.base_type.__name__}"
                )
            if key in self._registry:
                raise RuntimeError(f"{self.name}: '{key}' already registered")
            self._registry[key] = cls
            return cls

        return decorator

    def get(self, key: str) -> type[BaseModelT]:
        if key not in self._registry:
            raise ValueError(
                f"Unknown {self.name} '{key}'. Available: {list(self._registry.keys())}"
            )
        return self._registry[key]

    def registered(self) -> Mapping[str, type[BaseModelT]]:
        return MappingProxyType(self._registry)

    def load(self, value: Any) -> BaseModelT:
        if isinstance(value, self.base_type):
            return value

        if not tg.is_dictionary(value):
            raise TypeError(
                f"Expected a dict or a {self.base_type.__name__} instance, got {type(value)}"
            )

        if self.discriminator not in value:
            raise ValueError(
                f"Missing '{self.discriminator}' field for {self.base_type.__name__}"
            )
        key = value.pop(self.discriminator)

        cls = self.get(key)
        return cls.model_validate(value)

    @property
    def Validator(self) -> BeforeValidator:
        return BeforeValidator(self.load)



import os
from abc import ABC, abstractmethod
from random import Random
from typing import Annotated, Any

import joblib
import pandas as pd
from numpy.random import default_rng
from pydantic import Field
from tqdm import tqdm

import phylogenie.generators.configs as cfg
from phylogenie.generators.factories import context
from phylogenie.utils import Registry

DATA_DIRNAME = "data"
METADATA_FILENAME = "metadata.csv"


class DatasetGenerator(ABC, cfg.StrictBaseModel):
    output_dir: str = "phylogenie-outputs"
    n_samples: int | dict[str, int] = 1
    n_jobs: int = -1
    seed: int | None = None
    context: cfg.Context = Field(default_factory=dict)

    @abstractmethod
    def _generate_from_context(
        self, data_dir: str, file_id: str, context: dict[str, Any], seed: int
    ) -> dict[str, Any]: ...

    def _generate_one(self, data_dir: str, file_id: str, seed: int) -> dict[str, Any]:
        rng = default_rng(seed)
        while True:
            try:
                ctx = context(self.context, rng)
                metadata = self._generate_from_context(data_dir, file_id, ctx, seed)
                return {"file_id": file_id, **ctx, **metadata}
            except TimeoutError:
                print("Simulation timed out. Retrying with different context...")

    def _generate(self, n_samples: int, output_dir: str, rng: Random):
        if os.path.exists(output_dir):
            print(f"Output directory {output_dir} already exists. Skipping.")
            return

        data_dir = os.path.join(output_dir, DATA_DIRNAME)
        os.makedirs(data_dir)

        jobs = joblib.Parallel(n_jobs=self.n_jobs, return_as="generator_unordered")(
            joblib.delayed(self._generate_one)(data_dir, str(i), rng.getrandbits(32))
            for i in range(n_samples)
        )
        df = pd.DataFrame(
            [j for j in tqdm(jobs, f"Generating {data_dir}...", n_samples)]
        )
        df.to_csv(os.path.join(output_dir, METADATA_FILENAME), index=False)

    def generate(self):
        rng = Random(self.seed)
        if isinstance(self.n_samples, dict):
            for key, n_samples in self.n_samples.items():
                output_dir = os.path.join(self.output_dir, key)
                self._generate(n_samples, output_dir, rng)
        else:
            self._generate(self.n_samples, self.output_dir, rng)


DatasetGeneratorRegistry: Registry["DatasetGenerator"] = Registry(
    DatasetGenerator, "data_type"
)
DatasetGeneratorConfig = Annotated[
    "DatasetGenerator", DatasetGeneratorRegistry.Validator
]
